#!/usr/bin/ruby -w
# Encoding: UTF-8
# frozen_string_literal: true
# =========================================================================== #
# require 'extracter/class/extract_this_archive.rb'
# =========================================================================== #
require 'extracter/base/base.rb'

module Extracter

class Extracter < ::Extracter::Base # === Extracter::Extracter

  # ========================================================================= #
  # === consider_verbosity_for
  # ========================================================================= #
  def consider_verbosity_for(i)
    i = i.dup
    unless be_verbose?
      case i
      # "tar -xvf" must become "tar -xvf" here.
      when COMMAND_TO_EXTRACT_TAR_XZ_FILES
        i.delete!('v')
      end
    end
    return i
  end

  # ========================================================================= #
  # === extract_this_archive                                    (extract tag)
  # ========================================================================= #
  def extract_this_archive(
      i, extract_to = extract_to?
    )
    i = File.absolute_path(i)
    # ======================================================================= #
    # Next handle the situation when we are on Windows:
    # ======================================================================= #
    if ::Extracter.are_we_on_windows?
      # ===================================================================== #
      # On windows we will overrule this completely, for now.
      # ===================================================================== #
      _ = UNPACK_COMMAND_TO_USE_ON_WINDOWS+' '+
          File.absolute_path(i)+
          ' | '+
          SECOND_UNPACK_COMMAND_TO_USE_ON_WINDOWS
      esystem _
      return
    end
    # ======================================================================= #
    # First determine whether we can extract the archive or whether we can
    # not:
    # ======================================================================= #
    if ::Extracter.is_this_a_valid_archive?(i) or i.end_with?('.pdf')
      if be_verbose?
        if show_only_the_short_name_of_the_archive?
          name_of_the_archive = File.basename(i)
          copn; e "Extracting `#{sfancy(name_of_the_archive)}` "\
                  "to `#{sdir(extract_to)}` next."
        else
          copn; e "Extracting `#{sfancy(i)}` to `#{sdir(extract_to)}` "\
                  "next."
        end
      end
      unless File.writable? extract_to
        copn; ewarn 'You do not have sufficient permissions to'
        copn; ewarn "modify #{sdir(extract_to)}."
        return
      end
      # ===================================================================== #
      # Next, pad it if it includes a ' ' character or (). This was
      # disabled as of July 2022.
      # ===================================================================== #
      # i = pad(i) if i.include?(' ') or i.include?(')')
      case i # case tag; those listed on top are more important.
      # ===================================================================== #
      # === .rpm
      #
      # This entry point will handle .rpm files.
      # ===================================================================== #
      when /\.rpm$/i
        name_of_the_directory = i.delete_suffix('.rpm')
        mkdir(name_of_the_directory)
        set_extract_to(File.absolute_path(name_of_the_directory))
        cd(name_of_the_directory)
        esystem 'rpm2cpio ../'+File.basename(i)+' | cpio -idmv'
        return # Early return.
      # ===================================================================== #
      # === .tar.xz
      #
      # Note that .txz is just .tar.xz. Support for .txz was added here
      # in January of 2012.
      # ===================================================================== #
      when /\.tar\.xz$/i,
           /\.txz$/i,
           /\.xz$/i
        esystem consider_verbosity_for(COMMAND_TO_EXTRACT_TAR_XZ_FILES)+' '+i+padded_extract_to?
      # ===================================================================== #
      # === .tar
      #
      # This entry point is for simple .tar files.
      # ===================================================================== #
      when /\.tar$/i
        esystem COMMAND_TO_EXTRACT_TAR_FILES+' '+i+padded_extract_to?
      # ===================================================================== #
      # === zip
      # ===================================================================== #
      when /.zip$/i,
           /.xpi$/i,
           /.docx$/i,
           /.odt$/i, # .docx and .odt format types can be unpacked with zip too.
           /.apkg$/
        # =================================================================== #
        # 'ar -jxf' # unzip #{what} <-- this should work as well.
        # =================================================================== #
        i = pad(i) if i.include?(' ') or i.include?(')')
        esystem "unzip #{i}"
      # ===================================================================== #
      # === tgz
      #
      # This entry point will also handle ".tar.Z" files.
      # ===================================================================== #
      when /\.?tgz$/i,
           /\.?tar.Z$/i,
           /\.?taz$/i
        esystem COMMAND_TO_EXTRACT_TGZ_FILES+' '+i+padded_extract_to?
      # ===================================================================== #
      # === 7z
      # ===================================================================== #
      when '.7z' # 7z does not accept the -C commandline.
        # _ << '7za e' # <- Deprecated this variant as of 05.06.2020.
        esystem "7z x #{i}"
      # ===================================================================== #
      # === .tar.bz2
      # ===================================================================== #
      when /\.tar\.bz2$/i,
           /\.tbz$/i
        esystem COMMAND_TO_EXTRACT_TAR_BZ2_FILES+' '+i+padded_extract_to?
      # ===================================================================== #
      # === gz
      # ===================================================================== #
      when /\.?gz$/i,
           /\.?apk$/i
        if i.include? '.tar' # Handle .tar.gz here.
          esystem 'tar -zxvf '+i+padded_extract_to?
        else
          esystem 'gunzip '+i
        end
      # ===================================================================== #
      # === .bz2
      # ===================================================================== #
      when /\.?bz2$/,
           /\.?tbz2$/
        if i.include? '.tar' # Treat it as a .tar file.
          esystem 'tar -vjxf '+i
        else
          esystem 'bunzip2 '+i
        end
      # ===================================================================== #
      # === rar
      # ===================================================================== #
      when '.rar'
        check_whether_rar_is_available
        esystem 'unrar e '+i
      # ===================================================================== #
      # === .zst
      #
      # This entry point is for e. g. "pymol-2.3.0-3-x86_64.pkg.tar.zst".
      # ===================================================================== #
      when '.zst','.tar.zst'
        esystem COMMAND_TO_EXTRACT_ZST_ARCHIVES+' '+i
      # ===================================================================== #
      # === deb
      #
      # We could use dpkg-deb too, such as via "dpkg-deb". But using "ar"
      # seems to be the better choice.
      # ===================================================================== #
      when /\.?deb$/i
        esystem COMMAND_TO_EXTRACT_DEB_FILES+' '+i
      # ===================================================================== #
      # === gem
      #
      # The gem commands needs a specific --target=DIRECTORY option.
      # ===================================================================== #
      when /\.?gem$/i
        esystem GEM_UNPACK_COMMAND+' '+i+" --target=#{extract_to}"
      # ===================================================================== #
      # === lzma
      # ===================================================================== #
      when '.lzma'
        esystem "#{COMMAND_TO_EXTRACT_LZMA_FILES} #{i}"
      # ===================================================================== #
      # === bin
      #
      # This entry point allows the user to handle .bin files. In this
      # context, "handling" means to simply run that file as-is.
      # ===================================================================== #
      when /\.?bin$/i 
        esystem("./#{i}")
      # ===================================================================== #
      # === iso
      # ===================================================================== #
      when /\.?iso$/i
        try_to_extract_this_iso_file(i)
        return
      # ===================================================================== #
      # === img
      #
      # Note that .img in this context refers to squafhs .img files.
      # ===================================================================== #
      when /\.?img$/i,
           /\.?squashfs$/i
        try_to_extract_this_img_file(i)
        return # Must return early in this case.
      # ===================================================================== #
      # === lz
      #
      # This entry point requires lzip to be installed.
      # ===================================================================== #
      when /\.?lz$/i,
           /\.?tar\.?lz$/i
        esystem("#{COMMAND_TO_EXTRACT_LZ_FILES} #{i}#{padded_extract_to?}")
      # ===================================================================== #
      # === mp4
      #
      # If it is a .mp4 file, we delegate to MultimediaParadise.extract_audio.
      #
      # Usage example:
      #
      #   rubyextract foobar.mp4
      #
      # ===================================================================== #
      when /\.?mp4$/i
        begin
          require 'multimedia_paradise/audio/extract_audio/extract_audio.rb'
        rescue LoadError; end
        if Object.const_defined? :MultimediaParadise
          MultimediaParadise.extract_audio(i)
        end
      # ===================================================================== #
      # === ps
      # ===================================================================== #
      when/\.?ps$/i
        esystem 'ps2ascii '+i+padded_extract_to?
      # ===================================================================== #
      # === .jar
      # ===================================================================== #
      when /\.?jar$/i
        esystem COMMAND_TO_EXTRACT_JAR_ARCHIVES+' '+i
      # ===================================================================== #
      # === rpm
      # ===================================================================== #
      when '.rpm'
        esystem "#{COMMAND_TO_EXTRACT_BSDTAR_ARCHIVES} #{i}"
      # ===================================================================== #
      # === sxz
      # ===================================================================== #
      when '.sxz'
        esystem "unsquashfs #{i}"
      # ===================================================================== #
      # === pdf
      #
      # For pdf-files we will tap into the pdf_paradise project, if it
      # is available.
      # ===================================================================== #
      when /\.?pdf$/
        begin
          require 'pdf_paradise/utility_scripts/convert_pdf_to_text.rb'
          _ = PdfParadise::ConvertPdfToText.new(i)
          e _.output_file?
        rescue LoadError => error
          e 'No, not there.'
          pp error # Show the error to the user here.
        end
        return # Must return early in this case.
      else # else tag. We did not find the extension type.
        notify_the_user_that_this_extension_has_not_been_registered_yet(i)
      end
    else
      if File.exist? i
        notify_the_user_that_this_extension_has_not_been_registered_yet(i)
      else
        opnn; e "No file exists at `#{sfile(i)}`."
      end
    end
  end; alias do_extract_to      extract_this_archive # === do_extract_to
       alias do_extract_what_to extract_this_archive # === do_extract_what_to
       alias start              extract_this_archive # === start

end; end